import typing
from abc import ABCMeta, abstractmethod as abstract_method
from collections import OrderedDict
from functools import cmp_to_key
from inspect import isclass as is_class

__all__ = [
    'BaseRegistry',
    'ClassRegistry',
    'BaseMutableRegistry',
    'RegistryKeyError',
    'SortedClassRegistry',
]


class RegistryKeyError(KeyError):
    """
    Used to differentiate a registry lookup from a standard KeyError.

    This is especially useful when a registry class expects to extract values
    from dicts to generate keys.
    """
    pass


class BaseRegistry(typing.Mapping, metaclass=ABCMeta):
    """
    Base functionality for registries.
    """

    def __contains__(self, key: typing.Hashable) -> bool:
        """
        Returns whether the specified key is registered.
        """
        try:
            # Use :py:meth:`get_class` instead of :py:meth:`__getitem__`, to
            # avoid creating a new instance unnecessarily (i.e., prevent
            # errors if the corresponding class' constructor requires
            # arguments).
            self.get_class(key)
        except RegistryKeyError:
            return False
        else:
            return True

    def __dir__(self) -> typing.List[typing.Hashable]:
        return list(self.keys())

    def __getitem__(self, key: typing.Hashable) -> object:
        """
        Shortcut for calling :py:meth:`get` with empty args/kwargs.
        """
        return self.get(key)

    def __iter__(self) -> typing.Generator[typing.Hashable, None, None]:
        """
        Returns a generator for iterating over registry keys, in the
        order that they were registered.
        """
        return self.keys()

    @abstract_method
    def __len__(self) -> int:
        """
        Returns the number of registered classes.
        """
        raise NotImplementedError(
            'Not implemented in {cls}.'.format(cls=type(self).__name__),
        )

    def __missing__(self, key) -> typing.Optional[object]:
        """
        Defines what to do when trying to access an unregistered key.

        Default behaviour is to throw a typed exception, but you could override
        this in a subclass, e.g., to return a default value.
        """
        raise RegistryKeyError(key)

    @abstract_method
    def get_class(self, key: typing.Hashable) -> type:
        """
        Returns the class associated with the specified key.
        """
        raise NotImplementedError(
            'Not implemented in {cls}.'.format(cls=type(self).__name__),
        )

    def get(self, key: typing.Hashable, *args, **kwargs) -> object:
        """
        Creates a new instance of the class matching the specified key.

        :param key:
            The corresponding load key.

        :param args:
            Positional arguments passed to class initializer.
            Ignored if the class registry was initialized with a null template
            function.

        :param kwargs:
            Keyword arguments passed to class initializer.
            Ignored if the class registry was initialized with a null template
            function.

        References:
          - :py:meth:`__init__`
        """
        return self.create_instance(self.get_class(key), *args, **kwargs)

    @staticmethod
    def gen_lookup_key(key: typing.Hashable) -> typing.Hashable:
        """
        Used by :py:meth:`get` to generate a lookup key.

        You may override this method in a subclass, for example if you need to
        support legacy aliases, etc.
        """
        return key

    @staticmethod
    def create_instance(class_: type, *args, **kwargs) -> object:
        """
        Prepares the return value for :py:meth:`get`.

        You may override this method in a subclass, if you want to customize
        the way new instances are created.

        :param class_:
            The requested class.

        :param args:
            Positional keywords passed to :py:meth:`get`.

        :param kwargs:
            Keyword arguments passed to :py:meth:`get`.
        """
        return class_(*args, **kwargs)

    @abstract_method
    def items(self) -> typing.Generator[
        typing.Tuple[typing.Hashable, type], None, None]:
        """
        Iterates over registered classes and their corresponding keys, in the
        order that they were registered.
        """
        raise NotImplementedError(
            'Not implemented in {cls}.'.format(cls=type(self).__name__),
        )

    def keys(self) -> typing.Generator[typing.Hashable, None, None]:
        """
        Returns a generator for iterating over registry keys, in the order that
        they were registered.
        """
        for item in self.items():
            yield item[0]

    def values(self) -> typing.Generator[type, None, None]:
        """
        Returns a generator for iterating over registered classes, in the order
        that they were registered.
        """
        for item in self.items():
            yield item[1]


class BaseMutableRegistry(BaseRegistry, typing.MutableMapping,
    metaclass=ABCMeta):
    """
    Extends :py:class:`BaseRegistry` with methods that can be used to modify
    the registered classes.
    """

    def __init__(self, attr_name: typing.Optional[str] = None) -> None:
        """
        :param attr_name:
            If provided, :py:meth:`register` will automatically detect the key
            to use when registering new classes.
        """
        super(BaseMutableRegistry, self).__init__()

        self.attr_name = attr_name

        # Map lookup keys to readable keys.
        # Only needed when :py:meth:`gen_lookup_key` is overridden, but I'm not
        # good enough at reflection black magic to figure out how to do that (:
        self._lookup_keys: typing.Dict[typing.Hashable, typing.Hashable] = {}

    def __delitem__(self, key: typing.Hashable) -> None:
        """
        Provides alternate syntax for un-registering a class.
        """
        self._unregister(self.gen_lookup_key(key))
        del self._lookup_keys[key]

    def __repr__(self) -> str:
        return '{type}({attr_name!r})'.format(
            attr_name=self.attr_name,
            type=type(self).__name__,
        )

    def __setitem__(self, key: typing.Hashable,
            class_: type) -> None:
        """
        Provides alternate syntax for registering a class.
        """
        lookup_key = self.gen_lookup_key(key)

        self._register(lookup_key, class_)
        self._lookup_keys[key] = lookup_key

    def register(self,
            key: typing.Union[typing.Hashable, type]) -> \
            typing.Callable[[type], type]:
        """
        Decorator that registers a class with the registry.

        Example::

           registry = ClassRegistry(attr_name='widget_type')

           @registry.register
           class CustomWidget(BaseWidget):
             widget_type = 'custom'
             ...

           # Override the registry key:
           @registry.register('premium')
           class AdvancedWidget(BaseWidget):
             ...

        :param key:
            The registry key to use for the registered class.
            Optional if the registry's :py:attr:`attr_name` is set.
        """
        # ``@register`` usage:
        if is_class(key):
            if self.attr_name:
                attr_key = getattr(key, self.attr_name)
                lookup_key = self.gen_lookup_key(attr_key)

                # Note that ``getattr`` will raise an AttributeError if the
                # class doesn't have the required attribute.
                self._register(lookup_key, key)
                self._lookup_keys[attr_key] = lookup_key

                return key
            else:
                raise ValueError(
                    'Attempting to register {cls} to {registry} via decorator,'
                    ' but `{registry}.attr_key` is not set.'.format(
                        cls=key.__name__,
                        registry=type(self).__name__,
                    )
                )

        # ``@register('some_attr')`` usage:
        def _decorator(cls: type) -> type:
            lookup_key = self.gen_lookup_key(key)

            self._register(lookup_key, cls)
            self._lookup_keys[key] = lookup_key

            return cls

        return _decorator

    def unregister(self, key: typing.Hashable) -> type:
        """
        Unregisters the class with the specified key.

        :param key:
            The registry key to remove (not the registered class!).

        :return:
            The class that was unregistered.

        :raise:
            - :py:class:`KeyError` if the key is not registered.
        """
        result = self._unregister(self.gen_lookup_key(key))
        del self._lookup_keys[key]

        return result

    @abstract_method
    def _register(self, key: typing.Hashable,
            class_: type) -> None:
        """
        Registers a class with the registry.

        :param key: Has already been processed by :py:meth:`gen_lookup_key`.
        """
        raise NotImplementedError(
            'Not implemented in {cls}.'.format(cls=type(self).__name__),
        )

    @abstract_method
    def _unregister(self, key: typing.Hashable) -> type:
        """
        Unregisters the class at the specified key.

        :param key: Has already been processed by :py:meth:`gen_lookup_key`.
        """
        raise NotImplementedError(
            'Not implemented in {cls}.'.format(cls=type(self).__name__),
        )


class ClassRegistry(BaseMutableRegistry):
    """
    Maintains a registry of classes and provides a generic factory for
    instantiating them.
    """

    def __init__(
            self,
            attr_name: typing.Optional[str] = None,
            unique: bool = False,
    ) -> None:
        """
        :param attr_name:
            If provided, :py:meth:`register` will automatically detect the key
            to use when registering new classes.

        :param unique:
            Determines what happens when two classes are registered with the
            same key:

            - ``True``: A :py:class:`KeyError` will be raised.
            - ``False``: The second class will replace the first one.
        """
        super(ClassRegistry, self).__init__(attr_name)

        self.unique = unique

        self._registry: typing.OrderedDict[
            typing.Hashable, type] = OrderedDict()

    def __len__(self) -> int:
        """
        Returns the number of registered classes.
        """
        return len(self._registry)

    def __repr__(self) -> str:
        return '{type}(attr_name={attr_name!r}, unique={unique!r})'.format(
            attr_name=self.attr_name,
            type=type(self).__name__,
            unique=self.unique,
        )

    def get_class(self, key: typing.Hashable) -> typing.Optional[type]:
        """
        Returns the class associated with the specified key.
        """
        lookup_key = self.gen_lookup_key(key)

        try:
            return self._registry[lookup_key]
        except KeyError:
            return self.__missing__(lookup_key)

    def items(self) -> typing.Generator[
        typing.Tuple[typing.Hashable, type], None, None]:
        """
        Iterates over all registered classes, in the order they were added.
        """
        for key, lookup_key in self._lookup_keys.items():
            yield key, self._registry[lookup_key]

    def _register(self, key: typing.Hashable,
            class_: type) -> None:
        """
        Registers a class with the registry.

        :param key: Has already been processed by :py:meth:`gen_lookup_key`.
        """
        if key in ['', None]:
            raise ValueError(
                'Attempting to register class {cls} '
                'with empty registry key {key!r}.'.format(
                    cls=class_.__name__,
                    key=key,
                ),
            )

        if self.unique and (key in self._registry):
            raise RegistryKeyError(
                '{cls} with key {key!r} is already registered.'.format(
                    cls=class_.__name__,
                    key=key,
                ),
            )

        self._registry[key] = class_

    def _unregister(self, key: typing.Hashable) -> type:
        """
        Unregisters the class at the specified key.

        :param key: Has already been processed by :py:meth:`gen_lookup_key`.
        """
        return (
            self._registry.pop(key)
            if key in self._registry
            else self.__missing__(key)
        )


class SortedClassRegistry(ClassRegistry):
    """
    A ClassRegistry that uses a function to determine sort order when
    iterating.
    """

    def __init__(
            self,
            sort_key: typing.Union[
                str,
                typing.Callable[
                    [
                        typing.Tuple[typing.Hashable, type, typing.Hashable],
                        typing.Tuple[typing.Hashable, type, typing.Hashable]
                    ],
                    int,
                ],
            ],
            attr_name: typing.Optional[str] = None,
            unique: bool = False,
            reverse: bool = False,
    ) -> None:
        """
        :param sort_key:
            Attribute name or callable, used to determine the sort value.

            If callable, must accept two tuples of (key, class, lookup_key).

        :param attr_name:
            If provided, :py:meth:`register` will automatically detect the key
            to use when registering new classes.

        :param unique:
            Determines what happens when two classes are registered with the
            same key:

            - ``True``: The second class will replace the first one.
            - ``False``: A ``ValueError`` will be raised.

        :param reverse:
            Whether to reverse the sort ordering.
        """
        super(SortedClassRegistry, self).__init__(attr_name, unique)

        self._sort_key = (
            sort_key
            if callable(sort_key)
            else self.create_sorter(sort_key)
        )

        self.reverse = reverse

    def items(self) -> typing.Generator[
        typing.Tuple[typing.Hashable, type], None, None]:
        for (key, class_, _) in sorted(
                # Provide human-readable key and lookup key to the sorter...
                ((key, class_, self.gen_lookup_key(key)) for
                        (key, class_) in super().items()),
                key=self._sort_key,
                reverse=self.reverse,
        ):
            # ... but for parity with other ClassRegistry types, only include
            # the human-readable key in the result.
            yield key, class_

    @staticmethod
    def create_sorter(sort_key: str) -> typing.Callable[
        [
            typing.Tuple[typing.Hashable, type],
            typing.Tuple[typing.Hashable, type]
        ],
        int,
    ]:
        """
        Given a sort key, creates a function that can be used to sort items
        when iterating over the registry.
        """

        def sorter(a, b):
            a_attr = getattr(a[1], sort_key)
            b_attr = getattr(b[1], sort_key)

            return (a_attr > b_attr) - (a_attr < b_attr)

        return cmp_to_key(sorter)



import typing

from .registry import BaseMutableRegistry, RegistryKeyError

__all__ = [
    'RegistryPatcher',
]


class RegistryPatcher(object):
    """
    Creates a context in which classes are temporarily registered with a class
    registry, then removed when the context exits.

    Note: only mutable registries can be patched!
    """

    class DoesNotExist(object):
        """
        Used to identify a value that did not exist before we started.
        """
        pass

    def __init__(self, registry: BaseMutableRegistry, *args, **kwargs) -> None:
        """
        :param registry:
            A :py:class:`MutableRegistry` instance to patch.

        :param args:
            Classes to add to the registry.

            This behaves the same as decorating each class with
            ``@registry.register``.

            Note: ``registry.attr_name`` must be set!

        :param kwargs:
            Same as ``args``, except you explicitly specify the registry keys.

            In the event of a conflict, values in ``args`` override values in
            ``kwargs``.
        """
        super(RegistryPatcher, self).__init__()

        for class_ in args:
            kwargs[getattr(class_, registry.attr_name)] = class_

        self.target = registry

        self._new_values = kwargs
        self._prev_values = {}

    def __enter__(self) -> None:
        self.apply()

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.restore()

    def apply(self) -> None:
        """
        Applies the new values.
        """
        # Back up previous values.
        self._prev_values = {
            key: self._get_value(key, self.DoesNotExist)
            for key in self._new_values
        }

        # Patch values.
        for key, value in self._new_values.items():
            # Remove the existing value first (prevents issues if the registry
            # has ``unique=True``).
            self._del_value(key)

            if value is not self.DoesNotExist:
                self._set_value(key, value)

    def restore(self) -> None:
        """
        Restores previous settings.
        """
        # Restore previous settings.
        for key, value in self._prev_values.items():
            # Remove the existing value first (prevents issues if the registry
            # has ``unique=True``).
            self._del_value(key)

            if value is not self.DoesNotExist:
                self._set_value(key, value)

    def _get_value(self, key: typing.Hashable, default=None) -> \
            typing.Optional[type]:
        try:
            return self.target.get_class(key)
        except RegistryKeyError:
            return default

    def _set_value(self, key: typing.Hashable, value: type) -> None:
        self.target.register(key)(value)

    def _del_value(self, key: typing.Hashable) -> None:
        try:
            self.target.unregister(key)
        except RegistryKeyError:
            pass



from abc import ABCMeta
from inspect import isabstract as is_abstract

from .registry import BaseMutableRegistry

__all__ = [
    'AutoRegister',
]


def AutoRegister(registry: BaseMutableRegistry,
        base_type: type = ABCMeta) -> type:
    """
    Creates a metaclass that automatically registers all non-abstract
    subclasses in the specified registry.

    IMPORTANT:  Python defines abstract as "having at least one unimplemented
    abstract method"; specifying :py:class:`ABCMeta` as the metaclass is not
    enough!

    Example::

       commands = ClassRegistry(attr_name='command_name')

       # Specify ``AutoRegister`` as the metaclass:
       class BaseCommand(metaclass=AutoRegister(commands)):
         @abstractmethod
         def print(self):
           raise NotImplementedError()

       class PrintCommand(BaseCommand):
         command_name = 'print'

         def print(self):
           ...

       print(list(commands.items())) # [('print', PrintCommand)]

    :param registry:
        The registry that new classes will be added to.

        Note: the registry's ``attr_name`` attribute must be set!

    :param base_type:
        The base type of the metaclass returned by this function.

        99.99% of the time, this should be :py:class:`ABCMeta`.
    """
    if not registry.attr_name:
        raise ValueError(
            'Missing `attr_name` in {registry}.'.format(registry=registry),
        )

    class _metaclass(base_type):
        def __init__(self, what, bases=None, attrs=None):
            super(_metaclass, self).__init__(what, bases, attrs)

            if not is_abstract(self):
                registry.register(self)

    return _metaclass



from importlib.metadata import entry_points

import typing

from . import BaseRegistry

__all__ = [
    'EntryPointClassRegistry',
]


class EntryPointClassRegistry(BaseRegistry):
    """
    A class registry that loads classes using setuptools entry points.
    """

    def __init__(
            self,
            group: str,
            attr_name: typing.Optional[str] = None,
    ) -> None:
        """
        :param group:
            The name of the entry point group that will be used to load new
            classes.

        :param attr_name:
            If set, the registry will "brand" each class with its corresponding
            registry key.  This makes it easier to perform reverse lookups
            later.

            Note: if a class already defines this attribute, the registry will
            overwrite it!
        """
        super(EntryPointClassRegistry, self).__init__()

        self.attr_name = attr_name
        self.group = group

        self._cache: typing.Optional[typing.Dict[typing.Hashable, type]] = None
        """
        Caches registered classes locally, so that we don't have to keep
        iterating over entry points.
        """

        # If :py:attr:`attr_name` is set, warm the cache immediately, to apply
        # branding.
        if self.attr_name:
            self._get_cache()

    def __len__(self) -> int:
        return len(self._get_cache())

    def __repr__(self) -> str:
        return '{type}(group={group!r})'.format(
            group=self.group,
            type=type(self).__name__,
        )

    def get(self, key: typing.Hashable, *args, **kwargs) -> object:
        instance = super(EntryPointClassRegistry, self).get(key, *args,
            **kwargs)

        if self.attr_name:
            # Apply branding to the instance explicitly.
            # This is particularly important if the corresponding entry point
            # references a function or method.
            setattr(instance, self.attr_name, key)

        return instance

    def get_class(self, key: typing.Hashable) -> typing.Optional[type]:
        try:
            cls = self._get_cache()[key]
        except KeyError:
            cls = self.__missing__(key)

        return cls

    def items(self) -> typing.ItemsView[typing.Hashable, type]:
        return self._get_cache().items()

    def refresh(self):
        """
        Purges the local cache.  The next access attempt will reload all entry
        points.

        This is useful if you load a distribution at runtime... such as during
        unit tests for class-registry.  Otherwise, it probably serves no useful
        purpose
        """
        self._cache = None

    def _get_cache(self) -> typing.Dict[typing.Hashable, type]:
        """
        Populates the cache (if necessary) and returns it.
        """
        if self._cache is None:
            self._cache = {}
            for e in entry_points(group=self.group):
                cls = e.load()

                # Try to apply branding, but only for compatible types (i.e.,
                # functions and methods can't be branded this way).
                if self.attr_name and isinstance(cls, type):
                    setattr(cls, self.attr_name, e.name)

                self._cache[e.name] = cls

        return self._cache



import typing
from collections import defaultdict

from . import ClassRegistry

__all__ = [
    'ClassRegistryInstanceCache',
]


class ClassRegistryInstanceCache(object):
    """
    Wraps a ClassRegistry instance, caching instances as they are created.

    This allows you to create [multiple] registries that cache INSTANCES
    locally (so that they can be scoped and garbage-collected), while keeping
    the CLASS registry separate.

    Note that the internal class registry is copied by reference, so any
    classes that are registered afterward are accessible to both the
    ClassRegistry and the ClassRegistryInstanceCache.
    """

    def __init__(self, class_registry: ClassRegistry, *args, **kwargs) -> None:
        """
        :param class_registry:
            The wrapped ClassRegistry.

        :param args:
            Positional arguments passed to the class registry's template when
            creating new instances.

        :param kwargs:
            Keyword arguments passed to the class registry's template function
            when creating new instances.
        """
        super(ClassRegistryInstanceCache, self).__init__()

        self._registry = class_registry
        self._cache: typing.Dict[typing.Hashable, type] = {}

        self._key_map: typing.Dict[typing.Hashable, list] = defaultdict(list)

        self._template_args = args
        self._template_kwargs = kwargs

    def __getitem__(self, key):
        """
        Returns the cached instance associated with the specified key.
        """
        instance_key = self.get_instance_key(key)

        if instance_key not in self._cache:
            class_key = self.get_class_key(key)

            # Map lookup keys to cache keys so that we can iterate over them in
            # the correct order.
            self._key_map[class_key].append(instance_key)

            self._cache[instance_key] = self._registry.get(
                class_key,
                *self._template_args,
                **self._template_kwargs
            )

        return self._cache[instance_key]

    def __iter__(self) -> typing.Generator[type, None, None]:
        """
        Returns a generator for iterating over cached instances, using the
        wrapped registry to determine sort order.

        If a key has not been accessed yet, it will not be included.
        """
        for lookup_key in self._registry.keys():
            for cache_key in self._key_map[lookup_key]:
                yield self._cache[cache_key]

    def warm_cache(self) -> None:
        """
        Warms up the cache, ensuring that an instance is created for every key
        currently in the registry.

        .. note::
           This method does not account for any classes that may be added to
           the wrapped ClassRegistry in the future.
        """
        for key in self._registry.keys():
            self.__getitem__(key)

    def get_instance_key(self, key: typing.Hashable) -> typing.Hashable:
        """
        Generates a key that can be used to store/lookup values in the instance
        cache.

        :param key:
            Value provided to :py:meth:`__getitem__`.
        """
        return self.get_class_key(key)

    def get_class_key(self, key: typing.Hashable) -> typing.Hashable:
        """
        Generates a key that can be used to store/lookup values in the wrapped
        :py:class:`ClassRegistry` instance.

        This method is only invoked in the event of a cache miss.

        :param key:
            Value provided to :py:meth:`__getitem__`.
        """
        return self._registry.gen_lookup_key(key)
